"""OpenAI GPT provider."""

import json
from typing import Any

from miu_core.models import (
    Message,
    Response,
    TextContent,
    ToolUseContent,
    Usage,
)
from miu_core.providers.base import LLMProvider, ToolSchema

try:
    from openai import AsyncOpenAI
except ImportError as e:
    raise ImportError("openai package required. Install with: uv add miu-core[openai]") from e


class OpenAIProvider(LLMProvider):
    """OpenAI GPT LLM provider."""

    name = "openai"

    def __init__(self, model: str = "gpt-4o") -> None:
        self.model = model
        self._client = AsyncOpenAI()

    async def complete(
        self,
        messages: list[Message],
        tools: list[ToolSchema] | list[dict[str, Any]] | None = None,
        system: str | None = None,
        max_tokens: int = 4096,
    ) -> Response:
        """Send messages to OpenAI and get response."""
        api_messages = self._convert_messages(messages, system)

        kwargs: dict[str, Any] = {
            "model": self.model,
            "messages": api_messages,
            "max_tokens": max_tokens,
        }
        if tools:
            kwargs["tools"] = self._convert_tools(tools)

        response = await self._client.chat.completions.create(**kwargs)
        return self._convert_response(response)

    def _convert_messages(
        self, messages: list[Message], system: str | None
    ) -> list[dict[str, Any]]:
        """Convert internal messages to OpenAI format."""
        result: list[dict[str, Any]] = []

        if system:
            result.append({"role": "system", "content": system})

        for msg in messages:
            if msg.role == "system":
                result.append({"role": "system", "content": msg.get_text()})
            elif isinstance(msg.content, str):
                result.append({"role": msg.role, "content": msg.content})
            else:
                # Handle content blocks
                for block in msg.content:
                    if isinstance(block, TextContent):
                        result.append({"role": msg.role, "content": block.text})
                    elif isinstance(block, ToolUseContent):
                        # OpenAI uses tool_calls in assistant message
                        result.append(
                            {
                                "role": "assistant",
                                "tool_calls": [
                                    {
                                        "id": block.id,
                                        "type": "function",
                                        "function": {
                                            "name": block.name,
                                            "arguments": str(block.input),
                                        },
                                    }
                                ],
                            }
                        )
                    else:
                        # Tool result
                        result.append(
                            {
                                "role": "tool",
                                "tool_call_id": block.tool_use_id,
                                "content": block.content,
                            }
                        )

        return result

    def _convert_tools(
        self, tools: list[ToolSchema] | list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Convert tool schemas to OpenAI format."""
        return [
            {
                "type": "function",
                "function": {
                    "name": tool["name"],
                    "description": tool.get("description", ""),
                    "parameters": tool.get("input_schema", {}),
                },
            }
            for tool in tools
        ]

    def _convert_response(self, response: Any) -> Response:
        """Convert OpenAI response to internal format."""
        choice = response.choices[0]
        message = choice.message
        content: list[TextContent | ToolUseContent] = []

        if message.content:
            content.append(TextContent(text=message.content))

        if message.tool_calls:
            for tool_call in message.tool_calls:
                # Safely parse tool arguments
                tool_input: dict[str, Any] = {}
                if tool_call.function.arguments:
                    try:
                        tool_input = json.loads(tool_call.function.arguments)
                    except json.JSONDecodeError:
                        tool_input = {}
                content.append(
                    ToolUseContent(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        input=tool_input,
                    )
                )

        stop_reason = "end_turn"
        if choice.finish_reason == "tool_calls":
            stop_reason = "tool_use"
        elif choice.finish_reason == "stop":
            stop_reason = "end_turn"

        return Response(
            id=response.id,
            content=content,
            stop_reason=stop_reason,
            usage=Usage(
                input_tokens=response.usage.prompt_tokens if response.usage else 0,
                output_tokens=response.usage.completion_tokens if response.usage else 0,
            ),
        )



"""OpenTelemetry integration for miu-core.

This module provides OpenTelemetry tracing setup and utilities.
Requires: pip install miu-core[tracing]
"""

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased

from miu_core.tracing.types import TracingConfig

# Global tracer provider
_tracer_provider: TracerProvider | None = None


def setup_tracing(config: TracingConfig | None = None) -> trace.Tracer:
    """Setup OpenTelemetry tracing with OTLP exporter.

    Args:
        config: Tracing configuration. Uses defaults if not provided.

    Returns:
        Configured tracer instance.
    """
    global _tracer_provider

    if config is None:
        config = TracingConfig()

    if not config.enabled:
        # Return no-op tracer if disabled
        return trace.get_tracer(config.service_name)

    # Create resource with service info
    resource = Resource.create(
        {
            "service.name": config.service_name,
            "service.version": "0.1.0",
        }
    )

    # Create sampler based on sample rate
    sampler = TraceIdRatioBased(config.sample_rate)

    # Create provider
    _tracer_provider = TracerProvider(resource=resource, sampler=sampler)

    # Add OTLP exporter if endpoint is configured
    if config.endpoint:
        from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
            OTLPSpanExporter,
        )

        otlp_exporter = OTLPSpanExporter(endpoint=config.endpoint)
        _tracer_provider.add_span_processor(BatchSpanProcessor(otlp_exporter))

    # Add console exporter if enabled (for debugging)
    if config.console_export:
        _tracer_provider.add_span_processor(BatchSpanProcessor(ConsoleSpanExporter()))

    # Set global tracer provider
    trace.set_tracer_provider(_tracer_provider)

    return trace.get_tracer(config.service_name)


def get_tracer(name: str = "miu") -> trace.Tracer:
    """Get a tracer instance.

    If setup_tracing() hasn't been called, returns a no-op tracer.

    Args:
        name: Tracer name (defaults to "miu")

    Returns:
        Tracer instance
    """
    return trace.get_tracer(name)


def shutdown() -> None:
    """Shutdown the tracer provider and flush pending spans."""
    global _tracer_provider
    if _tracer_provider is not None:
        _tracer_provider.shutdown()
        _tracer_provider = None



"""Command registry for managing slash commands."""

from collections.abc import Iterator
from pathlib import Path

from miu_core.commands.schema import Command, parse_command_file


class CommandRegistry:
    """Registry for managing slash commands."""

    def __init__(self) -> None:
        self._commands: dict[str, Command] = {}

    def register(self, command: Command) -> None:
        """Register a command.

        Args:
            command: Command to register
        """
        self._commands[command.name] = command

    def get(self, name: str) -> Command | None:
        """Get a command by name.

        Args:
            name: Command name (without leading /)

        Returns:
            Command if found, None otherwise
        """
        return self._commands.get(name)

    def load_from_file(self, path: Path) -> Command | None:
        """Load a command from a markdown file.

        Args:
            path: Path to command file

        Returns:
            Loaded command or None if failed
        """
        if not path.exists() or path.suffix != ".md":
            return None

        content = path.read_text(encoding="utf-8")
        frontmatter, body = parse_command_file(content)

        command = Command(
            name=path.stem,
            description=frontmatter.get("description", ""),
            argument_hint=frontmatter.get("argument-hint", ""),
            content=body,
            source_path=path,
        )
        self.register(command)
        return command

    def load_from_directory(self, path: Path) -> int:
        """Load all commands from a directory.

        Args:
            path: Directory path containing command files

        Returns:
            Number of commands loaded
        """
        if not path.is_dir():
            return 0

        count = 0
        for file_path in path.glob("*.md"):
            if self.load_from_file(file_path):
                count += 1
        return count

    def list_commands(self) -> list[Command]:
        """Get all registered commands.

        Returns:
            List of all commands
        """
        return list(self._commands.values())

    def __len__(self) -> int:
        return len(self._commands)

    def __iter__(self) -> Iterator[Command]:
        return iter(self._commands.values())

    def __contains__(self, name: str) -> bool:
        return name in self._commands



"""MCP stdio transport."""

import asyncio
import json
from typing import Any


class StdioTransport:
    """Transport for MCP communication over stdio."""

    def __init__(self, command: list[str]) -> None:
        """Initialize transport with server command.

        Args:
            command: Command to start the MCP server (e.g., ["npx", "@modelcontextprotocol/server-filesystem"])
        """
        self.command = command
        self._process: asyncio.subprocess.Process | None = None
        self._request_id = 0

    async def start(self) -> None:
        """Start the MCP server process."""
        self._process = await asyncio.create_subprocess_exec(
            *self.command,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

    async def stop(self) -> None:
        """Stop the MCP server process."""
        if self._process:
            self._process.terminate()
            await self._process.wait()
            self._process = None

    async def send(self, message: dict[str, Any]) -> dict[str, Any]:
        """Send a message and receive response.

        Args:
            message: Message dict to send

        Returns:
            Response dict from server
        """
        if not self._process or not self._process.stdin or not self._process.stdout:
            raise RuntimeError("Transport not started")

        # Assign request ID if not present
        if "id" not in message or message["id"] is None:
            self._request_id += 1
            message["id"] = self._request_id

        # Send message
        data = json.dumps(message) + "\n"
        self._process.stdin.write(data.encode())
        await self._process.stdin.drain()

        # Read response
        response_line = await self._process.stdout.readline()
        if not response_line:
            raise RuntimeError("No response from MCP server")

        return json.loads(response_line.decode())  # type: ignore[no-any-return]

    @property
    def is_running(self) -> bool:
        """Check if transport is running."""
        return self._process is not None and self._process.returncode is None



"""Skill registry for managing loaded skills."""

from pathlib import Path

from miu_core.skills.base import Skill
from miu_core.skills.loader import SkillLoader


class SkillRegistry:
    """Registry for managing and querying skills."""

    def __init__(self) -> None:
        self._skills: dict[str, Skill] = {}
        self._loader = SkillLoader()

    def register(self, skill: Skill) -> None:
        """Register a skill."""
        self._skills[skill.name.lower()] = skill

    def get(self, name: str) -> Skill | None:
        """Get skill by name."""
        return self._skills.get(name.lower())

    def load_from_dir(self, path: Path) -> int:
        """Load skills from directory.

        Expects structure:
            path/
                skill-name-1/
                    SKILL.md
                skill-name-2/
                    SKILL.md

        Returns number of skills loaded.
        """
        if not path.exists() or not path.is_dir():
            return 0

        count = 0
        for skill_dir in path.iterdir():
            if not skill_dir.is_dir():
                continue
            skill = self._loader.load(skill_dir)
            if skill:
                self.register(skill)
                count += 1
        return count

    def build_system_prompt(self) -> str:
        """Build system prompt fragment from all skills."""
        if not self._skills:
            return ""

        parts = ["# Active Skills\n"]
        for skill in self._skills.values():
            parts.append(skill.to_prompt())
        return "\n".join(parts)

    def list_skills(self) -> list[Skill]:
        """Get all registered skills."""
        return list(self._skills.values())

    def __len__(self) -> int:
        return len(self._skills)

    def __iter__(self):  # type: ignore[no-untyped-def]
        return iter(self._skills.values())



"""Tests for tool framework."""

import pytest
from pydantic import BaseModel

from miu_core.tools import Tool, ToolContext, ToolRegistry, ToolResult


class EchoInput(BaseModel):
    message: str


class EchoTool(Tool):
    name = "echo"
    description = "Echoes the input message"

    def get_input_schema(self) -> type[BaseModel]:
        return EchoInput

    async def execute(self, ctx: ToolContext, message: str, **kwargs: object) -> ToolResult:
        return ToolResult(output=f"Echo: {message}")


class FailingTool(Tool):
    name = "failing"
    description = "Always fails"

    def get_input_schema(self) -> type[BaseModel]:
        return EchoInput

    async def execute(self, ctx: ToolContext, **kwargs: object) -> ToolResult:
        raise ValueError("Intentional failure")


class TestToolResult:
    def test_success_result(self) -> None:
        result = ToolResult(output="Success!")
        assert result.success is True
        assert result.error is None

    def test_error_result(self) -> None:
        result = ToolResult(output="Failed", success=False, error="Something went wrong")
        assert result.success is False
        assert result.error == "Something went wrong"


class TestToolContext:
    def test_defaults(self) -> None:
        ctx = ToolContext()
        assert ctx.working_dir == "."
        assert ctx.session_id == "default"

    def test_custom_values(self) -> None:
        ctx = ToolContext(working_dir="/tmp", session_id="session_123")
        assert ctx.working_dir == "/tmp"
        assert ctx.session_id == "session_123"


class TestTool:
    def test_to_schema(self) -> None:
        tool = EchoTool()
        schema = tool.to_schema()
        assert schema["name"] == "echo"
        assert schema["description"] == "Echoes the input message"
        assert "message" in schema["input_schema"]["properties"]

    @pytest.mark.asyncio
    async def test_execute(self) -> None:
        tool = EchoTool()
        ctx = ToolContext()
        result = await tool.execute(ctx, message="Hello")
        assert result.output == "Echo: Hello"
        assert result.success is True


class TestToolRegistry:
    def test_register_and_get(self) -> None:
        registry = ToolRegistry()
        tool = EchoTool()
        registry.register(tool)
        assert registry.get("echo") is tool

    def test_get_unknown_tool(self) -> None:
        registry = ToolRegistry()
        assert registry.get("unknown") is None

    def test_get_schemas(self) -> None:
        registry = ToolRegistry()
        registry.register(EchoTool())
        schemas = registry.get_schemas()
        assert len(schemas) == 1
        assert schemas[0]["name"] == "echo"

    @pytest.mark.asyncio
    async def test_execute(self) -> None:
        registry = ToolRegistry()
        registry.register(EchoTool())
        ctx = ToolContext()
        result = await registry.execute("echo", ctx, message="Test")
        assert result.output == "Echo: Test"

    @pytest.mark.asyncio
    async def test_execute_unknown_tool(self) -> None:
        registry = ToolRegistry()
        ctx = ToolContext()
        result = await registry.execute("unknown", ctx)
        assert result.success is False
        assert "not found" in result.output.lower()

    @pytest.mark.asyncio
    async def test_execute_handles_exception(self) -> None:
        registry = ToolRegistry()
        registry.register(FailingTool())
        ctx = ToolContext()
        result = await registry.execute("failing", ctx, message="test")
        assert result.success is False
        assert "Intentional failure" in result.error or ""

    def test_len(self) -> None:
        registry = ToolRegistry()
        assert len(registry) == 0
        registry.register(EchoTool())
        assert len(registry) == 1

    def test_iter(self) -> None:
        registry = ToolRegistry()
        registry.register(EchoTool())
        tools = list(registry)
        assert len(tools) == 1
        assert tools[0].name == "echo"



"""MCP (Model Context Protocol) client implementation."""

from miu_core.mcp.client import MCPClient
from miu_core.mcp.protocol import (
    CallToolRequest,
    CallToolResponse,
    InitializeRequest,
    InitializeResponse,
    ListToolsRequest,
    ListToolsResponse,
    MCPMessage,
    MCPTool,
)

__all__ = [
    "CallToolRequest",
    "CallToolResponse",
    "InitializeRequest",
    "InitializeResponse",
    "ListToolsRequest",
    "ListToolsResponse",
    "MCPClient",
    "MCPMessage",
    "MCPTool",
]



"""OpenTelemetry tracing integration for miu-core.

Install with: pip install miu-core[tracing]
"""

from miu_core.tracing.noop import Tracer
from miu_core.tracing.types import SpanKind, TracingConfig

__all__ = ["SpanKind", "Tracer", "TracingConfig", "get_tracer", "setup_tracing"]


def get_tracer(name: str = "miu") -> Tracer:
    """Get a tracer instance.

    Returns a no-op tracer if OpenTelemetry is not installed.
    """
    try:
        from miu_core.tracing.otel import get_tracer as _get_tracer

        return _get_tracer(name)  # type: ignore[no-any-return]
    except ImportError:
        from miu_core.tracing.noop import NoOpTracer

        return NoOpTracer()


def setup_tracing(config: TracingConfig | None = None) -> object:
    """Setup OpenTelemetry tracing.

    Args:
        config: Tracing configuration. Uses defaults if not provided.

    Returns:
        Configured tracer instance.

    Raises:
        ImportError: If OpenTelemetry packages are not installed.
    """
    try:
        from miu_core.tracing.otel import setup_tracing as _setup

        return _setup(config)
    except ImportError as e:
        raise ImportError(
            "OpenTelemetry packages not installed. Install with: pip install miu-core[tracing]"
        ) from e



"""Tests for tracing module."""

from miu_core.tracing import get_tracer
from miu_core.tracing.noop import NoOpSpan, NoOpTracer
from miu_core.tracing.types import SpanAttributes, SpanKind, TracingConfig


def test_tracing_config_defaults() -> None:
    """Test TracingConfig has sensible defaults."""
    config = TracingConfig()
    assert config.service_name == "miu"
    assert config.enabled is True
    assert config.sample_rate == 1.0
    assert config.console_export is False
    assert config.endpoint is None


def test_tracing_config_custom() -> None:
    """Test TracingConfig with custom values."""
    config = TracingConfig(
        service_name="test-service",
        endpoint="http://localhost:4317",
        enabled=True,
        sample_rate=0.5,
        console_export=True,
    )
    assert config.service_name == "test-service"
    assert config.endpoint == "http://localhost:4317"
    assert config.sample_rate == 0.5


def test_span_kind_enum() -> None:
    """Test SpanKind enum values."""
    assert SpanKind.AGENT == "agent"
    assert SpanKind.PROVIDER == "provider"
    assert SpanKind.TOOL == "tool"
    assert SpanKind.MEMORY == "memory"
    assert SpanKind.MCP == "mcp"


def test_span_attributes() -> None:
    """Test SpanAttributes constants."""
    # Agent attributes
    assert SpanAttributes.AGENT_NAME == "miu.agent.name"
    assert SpanAttributes.AGENT_QUERY == "miu.agent.query"
    assert SpanAttributes.AGENT_ITERATIONS == "miu.agent.iterations"

    # Provider attributes
    assert SpanAttributes.PROVIDER_NAME == "miu.provider.name"
    assert SpanAttributes.PROVIDER_MODEL == "miu.provider.model"
    assert SpanAttributes.PROVIDER_TOKENS_INPUT == "miu.provider.tokens.input"
    assert SpanAttributes.PROVIDER_TOKENS_OUTPUT == "miu.provider.tokens.output"

    # Tool attributes
    assert SpanAttributes.TOOL_NAME == "miu.tool.name"
    assert SpanAttributes.TOOL_SUCCESS == "miu.tool.success"


def test_noop_span() -> None:
    """Test NoOpSpan does nothing but doesn't error."""
    span = NoOpSpan()

    # All methods should be no-ops
    span.set_attribute("key", "value")
    span.set_status(None)
    span.record_exception(Exception("test"))
    span.end()

    # Context manager should work
    with span as s:
        assert s is span


def test_noop_tracer() -> None:
    """Test NoOpTracer returns no-op spans."""
    tracer = NoOpTracer()

    # start_as_current_span context manager
    with tracer.start_as_current_span("test-span") as span:
        assert isinstance(span, NoOpSpan)
        span.set_attribute("key", "value")

    # start_span returns NoOpSpan
    span = tracer.start_span("another-span")
    assert isinstance(span, NoOpSpan)


def test_get_tracer_returns_noop_when_otel_not_installed() -> None:
    """Test get_tracer returns no-op tracer when OTel not installed."""
    # This test assumes OTel is not installed in base test environment
    # If OTel IS installed, this will return a real tracer which is also fine
    tracer = get_tracer("test")
    # Should be usable regardless
    with tracer.start_as_current_span("test") as span:
        span.set_attribute("test", "value")



"""Slash commands system."""

from miu_core.commands.executor import CommandExecutor
from miu_core.commands.registry import CommandRegistry
from miu_core.commands.schema import Command

__all__ = [
    "Command",
    "CommandExecutor",
    "CommandRegistry",
]



"""Skill directory loader."""

from pathlib import Path

from miu_core.skills.base import Skill
from miu_core.skills.manifest import SkillManifest


class SkillLoader:
    """Loads skills from directories."""

    def load(self, skill_dir: Path) -> Skill | None:
        """Load a skill from directory containing SKILL.md."""
        manifest_path = skill_dir / "SKILL.md"
        if not manifest_path.exists():
            return None

        content = manifest_path.read_text(encoding="utf-8")
        manifest = SkillManifest.parse(content, name=skill_dir.name)

        # Resolve script paths
        scripts: list[Path] = []
        for script_ref in manifest.scripts:
            # Handle "scripts/xxx.py" or just "xxx.py"
            if ":" in script_ref:
                script_ref = script_ref.split(":")[0].strip().strip("`")
            script_path = skill_dir / script_ref
            if script_path.exists():
                scripts.append(script_path)

        # Resolve resource paths
        resources: list[Path] = []
        for resource_ref in manifest.resources:
            if ":" in resource_ref:
                resource_ref = resource_ref.split(":")[0].strip().strip("`")
            resource_path = skill_dir / resource_ref
            if resource_path.exists():
                resources.append(resource_path)

        return Skill(
            name=manifest.name,
            description=manifest.description,
            instructions=manifest.instructions,
            scripts=scripts,
            resources=resources,
            path=skill_dir,
        )



"""Multi-agent patterns for orchestration, pipelines, and routing."""

from miu_core.patterns.orchestrator import Orchestrator, OrchestratorConfig, TaskResult
from miu_core.patterns.pipeline import Pipeline, PipelineConfig, PipelineStage
from miu_core.patterns.routing import Router, RouterConfig, RouteResult, RoutingRule

__all__ = [
    "Orchestrator",
    "OrchestratorConfig",
    "Pipeline",
    "PipelineConfig",
    "PipelineStage",
    "RouteResult",
    "Router",
    "RouterConfig",
    "RoutingRule",
    "TaskResult",
]



"""MCP client implementation."""

from typing import Any

from miu_core.mcp.protocol import (
    CallToolRequest,
    CallToolResponse,
    InitializeRequest,
    InitializeResponse,
    ListToolsRequest,
    ListToolsResponse,
    MCPTool,
)
from miu_core.mcp.stdio import StdioTransport


class MCPClient:
    """MCP client for connecting to external tool servers."""

    def __init__(self, server_command: list[str]) -> None:
        """Initialize MCP client.

        Args:
            server_command: Command to start the MCP server
        """
        self.transport = StdioTransport(server_command)
        self._tools: list[MCPTool] = []
        self._server_info: dict[str, Any] = {}
        self._connected = False

    async def connect(self) -> None:
        """Initialize connection with MCP server."""
        await self.transport.start()

        # Send initialize request
        init_request = InitializeRequest()
        response = await self.transport.send(init_request.model_dump(by_alias=True))

        if "result" in response:
            init_response = InitializeResponse.model_validate(response["result"])
            self._server_info = init_response.server_info.model_dump()

        # List available tools
        list_request = ListToolsRequest()
        response = await self.transport.send(list_request.model_dump(by_alias=True))

        if "result" in response:
            list_response = ListToolsResponse.model_validate(response["result"])
            self._tools = list_response.tools

        self._connected = True

    async def disconnect(self) -> None:
        """Disconnect from MCP server."""
        await self.transport.stop()
        self._connected = False

    async def call_tool(self, name: str, arguments: dict[str, Any] | None = None) -> str:
        """Call a tool on the MCP server.

        Args:
            name: Tool name
            arguments: Tool arguments

        Returns:
            Tool result text
        """
        if not self._connected:
            raise RuntimeError("Not connected to MCP server")

        request = CallToolRequest(name=name, arguments=arguments or {})
        response = await self.transport.send(request.model_dump(by_alias=True))

        if "error" in response:
            error = response["error"]
            return f"Error: {error.get('message', 'Unknown error')}"

        if "result" in response:
            result = CallToolResponse.model_validate(response["result"])
            if result.content:
                return result.content[0].text
            return ""

        return ""

    def get_tools(self) -> list[dict[str, Any]]:
        """Get available tools as miu schemas.

        Returns:
            List of tool schemas compatible with miu agent
        """
        return [self._convert_tool(tool) for tool in self._tools]

    def _convert_tool(self, mcp_tool: MCPTool) -> dict[str, Any]:
        """Convert MCP tool to miu tool schema.

        Args:
            mcp_tool: MCP tool definition

        Returns:
            Miu-compatible tool schema
        """
        return {
            "name": f"mcp_{mcp_tool.name}",
            "description": mcp_tool.description or f"MCP tool: {mcp_tool.name}",
            "input_schema": mcp_tool.input_schema.model_dump(),
        }

    @property
    def is_connected(self) -> bool:
        """Check if client is connected."""
        return self._connected

    @property
    def server_info(self) -> dict[str, Any]:
        """Get server information."""
        return self._server_info

    async def __aenter__(self) -> "MCPClient":
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, *args: object) -> None:
        """Async context manager exit."""
        await self.disconnect()



"""Hook event definitions."""

from enum import Enum
from typing import Any

from pydantic import BaseModel


class HookEvent(str, Enum):
    """Lifecycle events that can trigger hooks."""

    SESSION_START = "session_start"
    SESSION_END = "session_end"
    PRE_MESSAGE = "pre_message"
    POST_MESSAGE = "post_message"
    PRE_TOOL_USE = "pre_tool_use"
    POST_TOOL_USE = "post_tool_use"


class HookInput(BaseModel):
    """Input passed to hook scripts."""

    event: HookEvent
    session_id: str | None = None
    tool_name: str | None = None
    tool_input: dict[str, Any] | None = None
    message: str | None = None
    extra: dict[str, Any] | None = None


class HookResult(BaseModel):
    """Result from hook execution."""

    success: bool = True
    output: str = ""
    should_block: bool = False
    modified_input: dict[str, Any] | None = None



"""Hook script executor."""

import asyncio
import sys
from pathlib import Path

from miu_core.hooks.events import HookInput, HookResult


class HookExecutor:
    """Execute hook scripts in various languages."""

    async def execute(self, script: Path, input_data: HookInput, timeout: int = 30) -> HookResult:
        """Execute a hook script with input data.

        Supports:
        - Python (.py)
        - Node.js (.js, .mjs)
        - Shell (.sh)
        """
        if not script.exists():
            return HookResult(
                success=False, output=f"Script not found: {script}", should_block=False
            )

        suffix = script.suffix.lower()
        if suffix == ".py":
            cmd = [sys.executable, str(script)]
        elif suffix in (".js", ".mjs"):
            cmd = ["node", str(script)]
        elif suffix == ".sh":
            cmd = ["bash", str(script)]
        else:
            return HookResult(
                success=False,
                output=f"Unsupported script type: {suffix}",
                should_block=False,
            )

        try:
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

            input_json = input_data.model_dump_json().encode()
            stdout, stderr = await asyncio.wait_for(proc.communicate(input_json), timeout=timeout)

            if proc.returncode != 0:
                return HookResult(
                    success=False,
                    output=stderr.decode("utf-8", errors="replace"),
                    should_block=False,
                )

            # Try to parse JSON output
            output_text = stdout.decode("utf-8", errors="replace").strip()
            try:
                return HookResult.model_validate_json(output_text)
            except Exception:
                # If not JSON, treat as plain text output
                return HookResult(success=True, output=output_text)

        except TimeoutError:
            if proc:
                proc.kill()
            return HookResult(
                success=False, output=f"Hook timed out after {timeout}s", should_block=False
            )
        except Exception as e:
            return HookResult(success=False, output=str(e), should_block=False)



"""Hooks system for lifecycle events."""

from miu_core.hooks.events import HookEvent, HookInput, HookResult
from miu_core.hooks.executor import HookExecutor
from miu_core.hooks.registry import HookRegistry

__all__ = ["HookEvent", "HookExecutor", "HookInput", "HookRegistry", "HookResult"]



"""miu-core: Core framework library for miu AI agent."""

from miu_core.version import get_version

__version__ = get_version()

# Lazy imports to avoid circular dependencies
__all__ = [
    "Orchestrator",
    "OrchestratorConfig",
    "Pipeline",
    "PipelineConfig",
    "Router",
    "RouterConfig",
    "__version__",
]


def __getattr__(name: str) -> object:
    """Lazy import patterns module."""
    if name in (
        "Orchestrator",
        "OrchestratorConfig",
        "Pipeline",
        "PipelineConfig",
        "Router",
        "RouterConfig",
    ):
        from miu_core import patterns

        return getattr(patterns, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Router pattern for directing requests to appropriate agents."""

import re
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel

from miu_core.agents.base import Agent
from miu_core.models import Response


class RouterConfig(BaseModel):
    """Configuration for router."""

    default_agent: str | None = None
    fallback_message: str = "Unable to route request to an appropriate agent."


@dataclass
class RoutingRule:
    """A rule for routing requests."""

    name: str
    agent: Agent
    condition: Callable[[str], bool]
    priority: int = 0
    metadata: dict[str, Any] | None = None


@dataclass
class RouteResult:
    """Result from routing a request."""

    agent_name: str
    response: Response
    matched_rule: str | None = None


class Router:
    """Routes requests to appropriate agents based on rules.

    The router evaluates rules in priority order and directs the request
    to the first matching agent. Rules can be keyword-based, regex-based,
    or use custom condition functions.

    Example:
        router = Router()

        # Add agents with keyword-based routing
        router.add_route(
            "code",
            code_agent,
            keywords=["code", "programming", "debug"]
        )
        router.add_route(
            "writing",
            writing_agent,
            keywords=["write", "essay", "article"]
        )
        router.add_route(
            "general",
            general_agent,
            condition=lambda q: True,  # Catch-all
            priority=-1
        )

        result = await router.route("Help me debug this Python code")
        # Routes to code_agent
    """

    def __init__(self, config: RouterConfig | None = None) -> None:
        self.config = config or RouterConfig()
        self._routes: list[RoutingRule] = []
        self._agents: dict[str, Agent] = {}

    def add_route(
        self,
        name: str,
        agent: Agent,
        *,
        keywords: list[str] | None = None,
        pattern: str | None = None,
        condition: Callable[[str], bool] | None = None,
        priority: int = 0,
        **metadata: Any,
    ) -> "Router":
        """Add a routing rule.

        Args:
            name: Route identifier
            agent: Agent to route to when rule matches
            keywords: List of keywords to match (case-insensitive)
            pattern: Regex pattern to match
            condition: Custom condition function receiving the query
            priority: Higher priority rules are checked first (default 0)
            **metadata: Additional metadata for the route

        Returns:
            Self for method chaining
        """
        self._agents[name] = agent

        # Build condition from keywords, pattern, or custom
        if condition:
            rule_condition = condition
        elif pattern:
            rule_condition = self._make_pattern_condition(pattern)
        elif keywords:
            rule_condition = self._make_keyword_condition(keywords)
        else:
            raise ValueError("Must provide keywords, pattern, or condition")

        self._routes.append(
            RoutingRule(
                name=name,
                agent=agent,
                condition=rule_condition,
                priority=priority,
                metadata=dict(metadata) if metadata else None,
            )
        )

        # Sort by priority (highest first)
        self._routes.sort(key=lambda r: -r.priority)
        return self

    async def route(self, query: str) -> RouteResult:
        """Route a query to the appropriate agent.

        Args:
            query: The user query to route

        Returns:
            RouteResult with the agent's response
        """
        # Check rules in priority order
        for rule in self._routes:
            if rule.condition(query):
                response = await rule.agent.run(query)
                return RouteResult(
                    agent_name=rule.name,
                    response=response,
                    matched_rule=rule.name,
                )

        # Fall back to default agent
        if self.config.default_agent and self.config.default_agent in self._agents:
            response = await self._agents[self.config.default_agent].run(query)
            return RouteResult(
                agent_name=self.config.default_agent,
                response=response,
                matched_rule=None,
            )

        raise ValueError(self.config.fallback_message)

    def get_route(self, query: str) -> str | None:
        """Get the route name without executing the agent.

        Args:
            query: The query to check routing for

        Returns:
            Name of the matched route or None
        """
        for rule in self._routes:
            if rule.condition(query):
                return rule.name

        return self.config.default_agent

    @property
    def routes(self) -> list[str]:
        """Return list of route names."""
        return [route.name for route in self._routes]

    @staticmethod
    def _make_pattern_condition(pattern: str) -> Callable[[str], bool]:
        """Create a regex-based condition."""
        compiled = re.compile(pattern, re.IGNORECASE)

        def condition(query: str) -> bool:
            return bool(compiled.search(query))

        return condition

    @staticmethod
    def _make_keyword_condition(keywords: list[str]) -> Callable[[str], bool]:
        """Create a keyword-based condition."""
        kw_lower = [k.lower() for k in keywords]

        def condition(query: str) -> bool:
            query_lower = query.lower()
            return any(k in query_lower for k in kw_lower)

        return condition



"""Skills framework for extending agent capabilities."""

from miu_core.skills.base import Skill
from miu_core.skills.loader import SkillLoader
from miu_core.skills.manifest import SkillManifest
from miu_core.skills.registry import SkillRegistry

__all__ = ["Skill", "SkillLoader", "SkillManifest", "SkillRegistry"]



"""LLM providers."""

from miu_core.providers.base import LLMProvider, ToolSchema

__all__ = ["LLMProvider", "ToolSchema", "create_provider"]


def create_provider(spec: str) -> LLMProvider:
    """Create provider from 'provider:model' spec.

    Examples:
        - anthropic:claude-sonnet-4-20250514
        - openai:gpt-4o
        - google:gemini-2.0-flash

    If no model specified, uses provider default.
    """
    if ":" in spec:
        provider_name, model = spec.split(":", 1)
    else:
        provider_name = spec
        model = None

    if provider_name == "anthropic":
        from miu_core.providers.anthropic import AnthropicProvider

        return AnthropicProvider(model=model) if model else AnthropicProvider()
    elif provider_name == "openai":
        from miu_core.providers.openai import OpenAIProvider

        return OpenAIProvider(model=model) if model else OpenAIProvider()
    elif provider_name == "google":
        from miu_core.providers.google import GoogleProvider

        return GoogleProvider(model=model) if model else GoogleProvider()
    else:
        raise ValueError(f"Unknown provider: {provider_name}")


# Lazy imports for optional providers
def __getattr__(name: str) -> type:
    if name == "AnthropicProvider":
        from miu_core.providers.anthropic import AnthropicProvider

        return AnthropicProvider
    if name == "OpenAIProvider":
        from miu_core.providers.openai import OpenAIProvider

        return OpenAIProvider
    if name == "GoogleProvider":
        from miu_core.providers.google import GoogleProvider

        return GoogleProvider
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Message and content models."""

from miu_core.models.messages import (
    ContentBlock,
    Message,
    Response,
    TextContent,
    ToolResultContent,
    ToolUseContent,
    Usage,
)

__all__ = [
    "ContentBlock",
    "Message",
    "Response",
    "TextContent",
    "ToolResultContent",
    "ToolUseContent",
    "Usage",
]



"""Short-term memory implementation."""

from miu_core.memory.base import Memory
from miu_core.memory.truncation import TruncationStrategy, truncate_fifo, truncate_sliding
from miu_core.models import Message


class ShortTermMemory(Memory):
    """In-memory conversation storage with truncation support."""

    def __init__(
        self,
        strategy: TruncationStrategy = TruncationStrategy.FIFO,
        max_messages: int = 100,
    ) -> None:
        self._messages: list[Message] = []
        self.strategy = strategy
        self.max_messages = max_messages

    @property
    def messages(self) -> list[Message]:
        """Get all messages."""
        return self._messages.copy()

    def add(self, message: Message) -> None:
        """Add a message to memory."""
        self._messages.append(message)

        # Auto-truncate if too many messages
        if len(self._messages) > self.max_messages:
            self._messages = self._messages[-self.max_messages :]

    def get_messages(self) -> list[Message]:
        """Get messages for LLM context."""
        return self._messages.copy()

    def truncate(self, max_tokens: int) -> int:
        """Truncate memory to fit within token limit."""
        if self.strategy == TruncationStrategy.FIFO:
            self._messages, removed = truncate_fifo(self._messages, max_tokens)
        elif self.strategy == TruncationStrategy.SLIDING:
            self._messages, removed = truncate_sliding(self._messages, max_tokens)
        else:
            removed = 0
        return removed

    def clear(self) -> None:
        """Clear all messages."""
        self._messages.clear()



"""Pipeline pattern for sequential agent processing."""

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

from pydantic import BaseModel

from miu_core.agents.base import Agent
from miu_core.models import Response


class PipelineConfig(BaseModel):
    """Configuration for pipeline."""

    stop_on_error: bool = True


@dataclass
class PipelineStage:
    """A stage in the pipeline."""

    name: str
    agent: Agent
    transform: Callable[[str, Response], str] | None = None
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class PipelineResult:
    """Result from pipeline execution."""

    success: bool
    stages_completed: int
    final_response: Response | None
    stage_responses: dict[str, Response]
    error: str | None = None
    failed_stage: str | None = None


class Pipeline:
    """Sequential agent chain where each stage's output feeds to the next.

    Pipelines are useful for tasks that require multiple processing steps,
    such as research -> summarize -> write -> edit.

    Example:
        pipeline = Pipeline()
        pipeline.add_stage("research", research_agent)
        pipeline.add_stage(
            "summarize",
            summarizer_agent,
            transform=lambda query, resp: f"Summarize: {resp.get_text()}"
        )
        pipeline.add_stage(
            "edit",
            editor_agent,
            transform=lambda query, resp: f"Edit this: {resp.get_text()}"
        )

        result = await pipeline.run("Research quantum computing")
    """

    def __init__(self, config: PipelineConfig | None = None) -> None:
        self.config = config or PipelineConfig()
        self._stages: list[PipelineStage] = []

    def add_stage(
        self,
        name: str,
        agent: Agent,
        transform: Callable[[str, Response], str] | None = None,
        **metadata: Any,
    ) -> "Pipeline":
        """Add a stage to the pipeline.

        Args:
            name: Stage identifier
            agent: Agent to execute at this stage
            transform: Optional function to transform previous response into
                       the next query. Receives (original_query, previous_response)
                       and returns the new query string.
            **metadata: Additional metadata for the stage

        Returns:
            Self for method chaining
        """
        self._stages.append(
            PipelineStage(
                name=name,
                agent=agent,
                transform=transform,
                metadata=dict(metadata),
            )
        )
        return self

    async def run(self, initial_query: str) -> PipelineResult:
        """Execute the pipeline sequentially.

        Args:
            initial_query: The query to start the pipeline with

        Returns:
            PipelineResult with all stage outputs
        """
        stage_responses: dict[str, Response] = {}
        current_query = initial_query
        last_response: Response | None = None

        for i, stage in enumerate(self._stages):
            # Transform query if transformer provided and not first stage
            if stage.transform and last_response:
                current_query = stage.transform(initial_query, last_response)

            try:
                response = await stage.agent.run(current_query)
                stage_responses[stage.name] = response
                last_response = response
                current_query = response.get_text()
            except Exception as e:
                if self.config.stop_on_error:
                    return PipelineResult(
                        success=False,
                        stages_completed=i,
                        final_response=last_response,
                        stage_responses=stage_responses,
                        error=str(e),
                        failed_stage=stage.name,
                    )
                # Continue without this stage's response
                continue

        return PipelineResult(
            success=True,
            stages_completed=len(self._stages),
            final_response=last_response,
            stage_responses=stage_responses,
        )

    def __len__(self) -> int:
        """Return number of stages."""
        return len(self._stages)

    @property
    def stages(self) -> list[str]:
        """Return list of stage names."""
        return [stage.name for stage in self._stages]



"""Tests for message models."""

from miu_core.models import (
    Message,
    Response,
    TextContent,
    ToolResultContent,
    ToolUseContent,
    Usage,
)


class TestTextContent:
    def test_create(self) -> None:
        content = TextContent(text="Hello")
        assert content.type == "text"
        assert content.text == "Hello"


class TestToolUseContent:
    def test_create(self) -> None:
        content = ToolUseContent(
            id="tool_123",
            name="read_file",
            input={"path": "/tmp/test.txt"},
        )
        assert content.type == "tool_use"
        assert content.id == "tool_123"
        assert content.name == "read_file"
        assert content.input == {"path": "/tmp/test.txt"}


class TestToolResultContent:
    def test_create(self) -> None:
        content = ToolResultContent(
            tool_use_id="tool_123",
            content="File contents here",
        )
        assert content.type == "tool_result"
        assert content.tool_use_id == "tool_123"
        assert content.is_error is False

    def test_error_result(self) -> None:
        content = ToolResultContent(
            tool_use_id="tool_123",
            content="File not found",
            is_error=True,
        )
        assert content.is_error is True


class TestMessage:
    def test_string_content(self) -> None:
        msg = Message(role="user", content="Hello")
        assert msg.get_text() == "Hello"

    def test_list_content(self) -> None:
        msg = Message(
            role="assistant",
            content=[
                TextContent(text="Line 1"),
                TextContent(text="Line 2"),
            ],
        )
        assert msg.get_text() == "Line 1\nLine 2"

    def test_mixed_content_extracts_text(self) -> None:
        msg = Message(
            role="assistant",
            content=[
                TextContent(text="Hello"),
                ToolUseContent(id="1", name="test", input={}),
            ],
        )
        assert msg.get_text() == "Hello"


class TestResponse:
    def test_get_text(self) -> None:
        response = Response(
            id="resp_123",
            content=[TextContent(text="Hello"), TextContent(text="World")],
        )
        assert response.get_text() == "Hello\nWorld"

    def test_get_tool_uses(self) -> None:
        tool_use = ToolUseContent(id="1", name="test", input={})
        response = Response(
            id="resp_123",
            content=[TextContent(text="Hello"), tool_use],
        )
        tool_uses = response.get_tool_uses()
        assert len(tool_uses) == 1
        assert tool_uses[0].name == "test"

    def test_usage(self) -> None:
        response = Response(
            id="resp_123",
            content=[],
            usage=Usage(input_tokens=100, output_tokens=50),
        )
        assert response.usage is not None
        assert response.usage.input_tokens == 100
        assert response.usage.output_tokens == 50



"""Version utilities for miu-core."""

from importlib.metadata import PackageNotFoundError, version
from typing import NamedTuple


class VersionInfo(NamedTuple):
    """Structured version information."""

    major: int
    minor: int
    patch: int
    prerelease: str = ""


def get_version() -> str:
    """Get package version from metadata or fallback."""
    try:
        return version("miu-core")
    except PackageNotFoundError:
        return "0.1.0"


def version_info() -> VersionInfo:
    """Parse version into components."""
    v = get_version()
    parts = v.replace("-", "+").split("+")
    version_str = parts[0]
    prerelease = parts[1] if len(parts) > 1 else ""

    nums = [int(x) for x in version_str.split(".")[:3]]
    while len(nums) < 3:
        nums.append(0)

    return VersionInfo(nums[0], nums[1], nums[2], prerelease)



"""No-op tracer implementation for when OpenTelemetry is not installed."""

from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Protocol


class Span(Protocol):
    """Protocol for span objects."""

    def set_attribute(self, key: str, value: Any) -> None: ...
    def set_status(self, status: Any) -> None: ...
    def record_exception(self, exception: BaseException) -> None: ...
    def end(self) -> None: ...
    def __enter__(self) -> "Span": ...
    def __exit__(self, *args: Any) -> None: ...


class Tracer(Protocol):
    """Protocol for tracer objects."""

    def start_as_current_span(self, name: str, **kwargs: Any) -> Any: ...
    def start_span(self, name: str, **kwargs: Any) -> Span: ...


class NoOpSpan:
    """No-op span that does nothing."""

    def set_attribute(self, key: str, value: Any) -> None:
        """No-op set attribute."""
        pass

    def set_status(self, status: Any) -> None:
        """No-op set status."""
        pass

    def record_exception(self, exception: BaseException) -> None:
        """No-op record exception."""
        pass

    def end(self) -> None:
        """No-op end."""
        pass

    def __enter__(self) -> "NoOpSpan":
        """Context manager enter."""
        return self

    def __exit__(self, *args: Any) -> None:
        """Context manager exit."""
        pass


class NoOpTracer:
    """No-op tracer that returns no-op spans."""

    @contextmanager
    def start_as_current_span(self, name: str, **kwargs: Any) -> Iterator[NoOpSpan]:
        """Return a no-op span context manager."""
        yield NoOpSpan()

    def start_span(self, name: str, **kwargs: Any) -> NoOpSpan:
        """Return a no-op span."""
        return NoOpSpan()



"""SKILL.md manifest parser."""

import re

from pydantic import BaseModel


class SkillManifest(BaseModel):
    """Parsed SKILL.md manifest."""

    name: str
    description: str = ""
    instructions: str = ""
    scripts: list[str] = []
    resources: list[str] = []

    @classmethod
    def parse(cls, content: str, name: str | None = None) -> "SkillManifest":
        """Parse SKILL.md content into manifest."""
        # Extract title (first # heading)
        title_match = re.search(r"^#\s+(.+)$", content, re.MULTILINE)
        skill_name = title_match.group(1).strip() if title_match else (name or "Unknown")

        # Extract sections
        sections: dict[str, str] = {}
        current_section = "preamble"
        current_content: list[str] = []

        for line in content.split("\n"):
            if line.startswith("## "):
                if current_content:
                    sections[current_section] = "\n".join(current_content).strip()
                current_section = line[3:].strip().lower()
                current_content = []
            else:
                current_content.append(line)

        if current_content:
            sections[current_section] = "\n".join(current_content).strip()

        # Parse lists from sections
        def parse_list(text: str) -> list[str]:
            items = []
            for line in text.split("\n"):
                match = re.match(r"^[-*]\s+(.+)$", line)
                if match:
                    items.append(match.group(1).strip())
            return items

        return cls(
            name=skill_name,
            description=sections.get("description", ""),
            instructions=sections.get("instructions", ""),
            scripts=parse_list(sections.get("scripts", "")),
            resources=parse_list(sections.get("resources", "")),
        )



"""Message and content block types."""

from typing import Any, Literal

from pydantic import BaseModel, Field


class TextContent(BaseModel):
    """Text content block."""

    type: Literal["text"] = "text"
    text: str


class ToolUseContent(BaseModel):
    """Tool use request from assistant."""

    type: Literal["tool_use"] = "tool_use"
    id: str
    name: str
    input: dict[str, Any]


class ToolResultContent(BaseModel):
    """Tool execution result."""

    type: Literal["tool_result"] = "tool_result"
    tool_use_id: str
    content: str
    is_error: bool = False


ContentBlock = TextContent | ToolUseContent | ToolResultContent


class Message(BaseModel):
    """Conversation message."""

    role: Literal["user", "assistant", "system"]
    content: str | list[ContentBlock]

    def get_text(self) -> str:
        """Extract text content from message."""
        if isinstance(self.content, str):
            return self.content
        texts = [block.text for block in self.content if isinstance(block, TextContent)]
        return "\n".join(texts)


class Usage(BaseModel):
    """Token usage information."""

    input_tokens: int = 0
    output_tokens: int = 0


class Response(BaseModel):
    """LLM response."""

    id: str
    content: list[ContentBlock] = Field(default_factory=list)
    stop_reason: str = "end_turn"
    usage: Usage | None = None

    def get_text(self) -> str:
        """Extract text content from response."""
        texts = [block.text for block in self.content if isinstance(block, TextContent)]
        return "\n".join(texts)

    def get_tool_uses(self) -> list[ToolUseContent]:
        """Extract tool use blocks from response."""
        return [block for block in self.content if isinstance(block, ToolUseContent)]



"""Google Gemini provider."""

import json
from typing import Any

from miu_core.models import (
    Message,
    Response,
    TextContent,
    ToolUseContent,
    Usage,
)
from miu_core.providers.base import LLMProvider, ToolSchema

try:
    from google import genai
    from google.genai import types
except ImportError as e:
    raise ImportError("google-genai package required. Install with: uv add miu-core[google]") from e


class GoogleProvider(LLMProvider):
    """Google Gemini LLM provider."""

    name = "google"

    def __init__(self, model: str = "gemini-2.0-flash") -> None:
        self.model = model
        self._client = genai.Client()

    async def complete(
        self,
        messages: list[Message],
        tools: list[ToolSchema] | list[dict[str, Any]] | None = None,
        system: str | None = None,
        max_tokens: int = 4096,
    ) -> Response:
        """Send messages to Gemini and get response."""
        contents = self._convert_messages(messages)

        config = types.GenerateContentConfig(
            max_output_tokens=max_tokens,
            system_instruction=system,
        )

        if tools:
            config.tools = self._convert_tools(tools)

        response = await self._client.aio.models.generate_content(
            model=self.model,
            contents=contents,
            config=config,
        )
        return self._convert_response(response)

    def _convert_messages(self, messages: list[Message]) -> list[types.Content]:
        """Convert internal messages to Gemini format."""
        result: list[types.Content] = []

        for msg in messages:
            if msg.role == "system":
                continue  # System handled via config

            role = "user" if msg.role == "user" else "model"

            if isinstance(msg.content, str):
                result.append(types.Content(role=role, parts=[types.Part(text=msg.content)]))
            else:
                parts: list[types.Part] = []
                for block in msg.content:
                    if isinstance(block, TextContent):
                        parts.append(types.Part(text=block.text))
                    elif isinstance(block, ToolUseContent):
                        parts.append(
                            types.Part(
                                function_call=types.FunctionCall(
                                    name=block.name,
                                    args=block.input,
                                )
                            )
                        )
                    else:
                        # Tool result
                        parts.append(
                            types.Part(
                                function_response=types.FunctionResponse(
                                    name="tool_result",
                                    response={"result": block.content},
                                )
                            )
                        )
                if parts:
                    result.append(types.Content(role=role, parts=parts))

        return result

    def _convert_tools(self, tools: list[ToolSchema] | list[dict[str, Any]]) -> list[types.Tool]:
        """Convert tool schemas to Gemini format."""
        declarations = []
        for tool in tools:
            schema = tool.get("input_schema", {})
            # Clean schema for Gemini
            clean_schema = self._clean_schema(schema)
            declarations.append(
                types.FunctionDeclaration(
                    name=tool["name"],
                    description=tool.get("description", ""),
                    parameters=clean_schema,
                )
            )
        return [types.Tool(function_declarations=declarations)]

    def _clean_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        """Clean JSON schema for Gemini compatibility."""
        # Remove unsupported keys
        clean = {k: v for k, v in schema.items() if k not in ("title", "$defs", "definitions")}
        if "properties" in clean:
            clean["properties"] = {k: self._clean_schema(v) for k, v in clean["properties"].items()}
        return clean

    def _convert_response(self, response: Any) -> Response:
        """Convert Gemini response to internal format."""
        content: list[TextContent | ToolUseContent] = []
        stop_reason = "end_turn"

        if response.candidates:
            candidate = response.candidates[0]
            for part in candidate.content.parts:
                if hasattr(part, "text") and part.text:
                    content.append(TextContent(text=part.text))
                elif hasattr(part, "function_call") and part.function_call:
                    fc = part.function_call
                    # Safely convert args to dict
                    args: dict[str, Any] = {}
                    if isinstance(fc.args, dict):
                        args = fc.args
                    elif fc.args:
                        try:
                            args = json.loads(str(fc.args))
                        except json.JSONDecodeError:
                            args = {}
                    content.append(
                        ToolUseContent(
                            id=f"call_{fc.name}",
                            name=fc.name,
                            input=args,
                        )
                    )
                    stop_reason = "tool_use"

        usage = None
        if hasattr(response, "usage_metadata") and response.usage_metadata:
            usage = Usage(
                input_tokens=response.usage_metadata.prompt_token_count or 0,
                output_tokens=response.usage_metadata.candidates_token_count or 0,
            )

        return Response(
            id=f"gemini_{id(response)}",
            content=content,
            stop_reason=stop_reason,
            usage=usage,
        )



"""Tracing types and configuration."""

from enum import Enum

from pydantic import BaseModel, Field


class SpanKind(str, Enum):
    """Types of spans for categorization."""

    AGENT = "agent"
    PROVIDER = "provider"
    TOOL = "tool"
    MEMORY = "memory"
    MCP = "mcp"


class TracingConfig(BaseModel):
    """Configuration for OpenTelemetry tracing."""

    service_name: str = Field(default="miu", description="Service name for traces")
    endpoint: str | None = Field(
        default=None,
        description="OTLP endpoint (e.g., http://localhost:4317). Uses env var if not set.",
    )
    enabled: bool = Field(default=True, description="Enable/disable tracing")
    sample_rate: float = Field(
        default=1.0, ge=0.0, le=1.0, description="Trace sample rate (0.0-1.0)"
    )
    console_export: bool = Field(
        default=False, description="Also export spans to console for debugging"
    )


# Semantic attribute names for miu spans
class SpanAttributes:
    """Standard attribute names for miu spans."""

    # Agent attributes
    AGENT_NAME = "miu.agent.name"
    AGENT_QUERY = "miu.agent.query"
    AGENT_ITERATIONS = "miu.agent.iterations"
    AGENT_MAX_ITERATIONS = "miu.agent.max_iterations"

    # Provider attributes
    PROVIDER_NAME = "miu.provider.name"
    PROVIDER_MODEL = "miu.provider.model"
    PROVIDER_TOKENS_INPUT = "miu.provider.tokens.input"
    PROVIDER_TOKENS_OUTPUT = "miu.provider.tokens.output"
    PROVIDER_STOP_REASON = "miu.provider.stop_reason"

    # Tool attributes
    TOOL_NAME = "miu.tool.name"
    TOOL_SUCCESS = "miu.tool.success"
    TOOL_ERROR = "miu.tool.error"

    # Memory attributes
    MEMORY_MESSAGE_COUNT = "miu.memory.message_count"
    MEMORY_TRUNCATED = "miu.memory.truncated"

    # Error attributes
    ERROR_TYPE = "miu.error.type"
    ERROR_MESSAGE = "miu.error.message"



"""Base memory interface."""

from abc import ABC, abstractmethod

from miu_core.models import Message


class Memory(ABC):
    """Abstract base class for memory implementations."""

    @property
    @abstractmethod
    def messages(self) -> list[Message]:
        """Get all messages in memory."""
        ...

    @abstractmethod
    def add(self, message: Message) -> None:
        """Add a message to memory."""
        ...

    @abstractmethod
    def get_messages(self) -> list[Message]:
        """Get messages for LLM context."""
        ...

    @abstractmethod
    def truncate(self, max_tokens: int) -> int:
        """Truncate memory to fit within token limit. Returns tokens removed."""
        ...

    @abstractmethod
    def clear(self) -> None:
        """Clear all messages from memory."""
        ...

    def __len__(self) -> int:
        """Get number of messages in memory."""
        return len(self.messages)



"""Tool registry for managing tools."""

from typing import Any

from miu_core.tools.base import Tool, ToolContext, ToolResult


class ToolRegistry:
    """Registry for managing and executing tools."""

    def __init__(self) -> None:
        self._tools: dict[str, Tool] = {}

    def register(self, tool: Tool) -> None:
        """Register a tool."""
        self._tools[tool.name] = tool

    def get(self, name: str) -> Tool | None:
        """Get tool by name."""
        return self._tools.get(name)

    def get_schemas(self) -> list[dict[str, Any]]:
        """Get all tool schemas for LLM."""
        return [tool.to_schema() for tool in self._tools.values()]

    async def execute(self, name: str, ctx: ToolContext, **kwargs: Any) -> ToolResult:
        """Execute a tool by name."""
        tool = self._tools.get(name)
        if not tool:
            return ToolResult(
                output=f"Tool not found: {name}",
                success=False,
                error=f"Unknown tool: {name}",
            )
        try:
            return await tool.execute(ctx, **kwargs)
        except Exception as e:
            return ToolResult(
                output=str(e),
                success=False,
                error=str(e),
            )

    def __len__(self) -> int:
        return len(self._tools)

    def __iter__(self) -> Any:
        return iter(self._tools.values())



"""Memory system for managing conversation history."""

from miu_core.memory.base import Memory
from miu_core.memory.short_term import ShortTermMemory
from miu_core.memory.truncation import TruncationStrategy

__all__ = ["Memory", "ShortTermMemory", "TruncationStrategy"]



"""Base LLM provider interface."""

from abc import ABC, abstractmethod
from typing import Any, TypedDict


class ToolSchema(TypedDict):
    """Tool definition schema for LLM."""

    name: str
    description: str
    input_schema: dict[str, Any]


class LLMProvider(ABC):
    """Abstract base class for LLM providers."""

    name: str
    model: str

    @abstractmethod
    async def complete(
        self,
        messages: list["Message"],
        tools: list[ToolSchema] | list[dict[str, Any]] | None = None,
        system: str | None = None,
        max_tokens: int = 4096,
    ) -> "Response":
        """Send messages to LLM and get response."""
        ...


# Import at end to avoid circular imports
from miu_core.models import Message, Response  # noqa: E402



"""Base agent interface."""

from abc import ABC, abstractmethod

from pydantic import BaseModel

from miu_core.memory import Memory as MemoryInterface
from miu_core.memory import ShortTermMemory
from miu_core.models import Response
from miu_core.providers.base import LLMProvider
from miu_core.tools.registry import ToolRegistry


class AgentConfig(BaseModel):
    """Agent configuration."""

    name: str = "agent"
    system_prompt: str = "You are a helpful AI assistant."
    max_iterations: int = 10
    max_context_tokens: int = 100000  # Auto-truncate at this limit


class Agent(ABC):
    """Abstract base class for agents."""

    def __init__(
        self,
        provider: LLMProvider,
        tools: ToolRegistry | None = None,
        config: AgentConfig | None = None,
        memory: MemoryInterface | None = None,
    ) -> None:
        self.provider = provider
        self.tools = tools or ToolRegistry()
        self.config = config or AgentConfig()
        self.memory = memory or ShortTermMemory()

    @abstractmethod
    async def run(self, query: str) -> Response:
        """Process a query and return response."""
        ...



"""Hook registry for managing hooks."""

from pathlib import Path

from miu_core.hooks.events import HookEvent, HookInput, HookResult
from miu_core.hooks.executor import HookExecutor


class HookRegistry:
    """Registry for managing and executing hooks."""

    def __init__(self) -> None:
        self._hooks: dict[HookEvent, list[Path]] = {event: [] for event in HookEvent}
        self._executor = HookExecutor()

    def register(self, event: HookEvent, script: Path) -> None:
        """Register a hook script for an event."""
        if script.exists():
            self._hooks[event].append(script)

    def load_from_dir(self, path: Path) -> int:
        """Load hooks from directory.

        Expected structure:
            path/
                session-start.py
                pre-tool-use.sh
                post-message.js
        """
        if not path.exists() or not path.is_dir():
            return 0

        # Map filenames to events
        name_map = {
            "session-start": HookEvent.SESSION_START,
            "session-end": HookEvent.SESSION_END,
            "pre-message": HookEvent.PRE_MESSAGE,
            "post-message": HookEvent.POST_MESSAGE,
            "pre-tool-use": HookEvent.PRE_TOOL_USE,
            "post-tool-use": HookEvent.POST_TOOL_USE,
        }

        count = 0
        for script in path.iterdir():
            if not script.is_file():
                continue
            stem = script.stem.lower()
            if stem in name_map:
                self.register(name_map[stem], script)
                count += 1
        return count

    async def execute(self, event: HookEvent, input_data: HookInput) -> list[HookResult]:
        """Execute all hooks for an event."""
        results: list[HookResult] = []
        for script in self._hooks[event]:
            result = await self._executor.execute(script, input_data)
            results.append(result)
            if result.should_block:
                break
        return results

    def has_hooks(self, event: HookEvent) -> bool:
        """Check if event has registered hooks."""
        return len(self._hooks[event]) > 0

    def __len__(self) -> int:
        return sum(len(hooks) for hooks in self._hooks.values())



"""MCP protocol types."""

from typing import Any

from pydantic import BaseModel, Field


class MCPMessage(BaseModel):
    """Base MCP message."""

    jsonrpc: str = "2.0"
    id: int | str | None = None


class ClientInfo(BaseModel):
    """MCP client information."""

    name: str
    version: str


class ServerInfo(BaseModel):
    """MCP server information."""

    name: str
    version: str


class MCPCapabilities(BaseModel):
    """MCP capabilities."""

    tools: dict[str, Any] | None = None
    resources: dict[str, Any] | None = None
    prompts: dict[str, Any] | None = None


class InitializeRequest(MCPMessage):
    """MCP initialize request."""

    method: str = "initialize"
    params: dict[str, Any] = Field(default_factory=dict)

    def __init__(
        self,
        protocol_version: str = "2024-11-05",
        capabilities: dict[str, Any] | None = None,
        client_info: ClientInfo | None = None,
        **kwargs: Any,
    ) -> None:
        params = {
            "protocolVersion": protocol_version,
            "capabilities": capabilities or {},
            "clientInfo": client_info.model_dump()
            if client_info
            else {"name": "miu", "version": "0.1.0"},
        }
        super().__init__(params=params, **kwargs)


class InitializeResponse(BaseModel):
    """MCP initialize response."""

    protocol_version: str = Field(alias="protocolVersion")
    capabilities: MCPCapabilities
    server_info: ServerInfo = Field(alias="serverInfo")


class MCPToolInputSchema(BaseModel):
    """MCP tool input schema."""

    type: str = "object"
    properties: dict[str, Any] = Field(default_factory=dict)
    required: list[str] = Field(default_factory=list)


class MCPTool(BaseModel):
    """MCP tool definition."""

    name: str
    description: str | None = None
    input_schema: MCPToolInputSchema = Field(
        alias="inputSchema", default_factory=MCPToolInputSchema
    )


class ListToolsRequest(MCPMessage):
    """MCP list tools request."""

    method: str = "tools/list"
    params: dict[str, Any] = Field(default_factory=dict)


class ListToolsResponse(BaseModel):
    """MCP list tools response."""

    tools: list[MCPTool]


class CallToolRequest(MCPMessage):
    """MCP call tool request."""

    method: str = "tools/call"
    params: dict[str, Any] = Field(default_factory=dict)

    def __init__(self, name: str, arguments: dict[str, Any] | None = None, **kwargs: Any) -> None:
        params = {"name": name, "arguments": arguments or {}}
        super().__init__(params=params, **kwargs)


class ToolContent(BaseModel):
    """MCP tool response content."""

    type: str = "text"
    text: str = ""


class CallToolResponse(BaseModel):
    """MCP call tool response."""

    content: list[ToolContent]
    is_error: bool = Field(default=False, alias="isError")



"""ReAct (Reasoning + Acting) agent implementation."""

from miu_core.agents.base import Agent, AgentConfig
from miu_core.memory import Memory
from miu_core.models import Message, Response, TextContent, ToolResultContent
from miu_core.providers.base import LLMProvider
from miu_core.tools.base import ToolContext
from miu_core.tools.registry import ToolRegistry
from miu_core.tracing import Tracer, get_tracer
from miu_core.tracing.types import SpanAttributes


class ReActAgent(Agent):
    """Agent implementing ReAct pattern."""

    def __init__(
        self,
        provider: LLMProvider,
        tools: ToolRegistry | None = None,
        config: AgentConfig | None = None,
        memory: Memory | None = None,
        working_dir: str = ".",
    ) -> None:
        super().__init__(provider, tools, config, memory)
        self.working_dir = working_dir
        self._tracer: Tracer = get_tracer("miu.agent")

    async def run(self, query: str) -> Response:
        """Execute ReAct loop for query."""
        with self._tracer.start_as_current_span("agent.run") as span:
            span.set_attribute(SpanAttributes.AGENT_NAME, self.config.name)
            span.set_attribute(SpanAttributes.AGENT_QUERY, query[:100])
            span.set_attribute(SpanAttributes.AGENT_MAX_ITERATIONS, self.config.max_iterations)

            self.memory.add(Message(role="user", content=query))

            # Auto-truncate if over context limit
            self.memory.truncate(self.config.max_context_tokens)

            iteration = 0
            for iteration in range(self.config.max_iterations):
                response = await self.provider.complete(
                    messages=self.memory.get_messages(),
                    tools=self.tools.get_schemas() if len(self.tools) > 0 else None,
                    system=self.config.system_prompt,
                )
                self.memory.add(Message(role="assistant", content=response.content))

                if response.stop_reason == "end_turn":
                    span.set_attribute(SpanAttributes.AGENT_ITERATIONS, iteration + 1)
                    return response

                if response.stop_reason == "tool_use":
                    await self._execute_tools(response)

            span.set_attribute(SpanAttributes.AGENT_ITERATIONS, iteration + 1)
            return Response(
                id="max_iterations",
                content=[TextContent(text="Maximum iterations reached")],
                stop_reason="max_iterations",
            )

    async def _execute_tools(self, response: Response) -> None:
        """Execute tool calls from response."""
        tool_uses = response.get_tool_uses()
        ctx = ToolContext(working_dir=self.working_dir)

        results: list[ToolResultContent] = []
        for tool_use in tool_uses:
            with self._tracer.start_as_current_span("tool.execute") as span:
                span.set_attribute(SpanAttributes.TOOL_NAME, tool_use.name)

                result = await self.tools.execute(tool_use.name, ctx, **tool_use.input)

                span.set_attribute(SpanAttributes.TOOL_SUCCESS, result.success)
                if not result.success and result.error:
                    span.set_attribute(SpanAttributes.TOOL_ERROR, result.error)

                results.append(
                    ToolResultContent(
                        tool_use_id=tool_use.id,
                        content=result.output,
                        is_error=not result.success,
                    )
                )

        self.memory.add(Message(role="user", content=results))



"""Command executor for parsing and expanding slash commands."""

from miu_core.commands.registry import CommandRegistry


class CommandExecutor:
    """Execute slash commands by parsing and expanding templates."""

    def __init__(self, registry: CommandRegistry) -> None:
        """Initialize executor with command registry.

        Args:
            registry: Registry containing available commands
        """
        self.registry = registry

    def parse(self, input_str: str) -> tuple[str, str] | None:
        """Parse '/command args' format.

        Args:
            input_str: Input string to parse

        Returns:
            Tuple of (command_name, args) or None if not a command
        """
        if not input_str.startswith("/"):
            return None

        stripped = input_str[1:].strip()
        if not stripped:
            return None

        parts = stripped.split(maxsplit=1)
        command_name = parts[0]
        args = parts[1] if len(parts) > 1 else ""

        return (command_name, args)

    def expand(self, command_name: str, args: str) -> str:
        """Expand command template with arguments.

        Args:
            command_name: Name of the command (without /)
            args: Arguments to pass to the command

        Returns:
            Expanded command content

        Raises:
            ValueError: If command not found
        """
        command = self.registry.get(command_name)
        if not command:
            raise ValueError(f"Unknown command: /{command_name}")
        return command.expand(args)

    def execute(self, input_str: str) -> str | None:
        """Parse and expand a command string.

        Args:
            input_str: Input string that may be a command

        Returns:
            Expanded command content or None if not a command

        Raises:
            ValueError: If command not found
        """
        parsed = self.parse(input_str)
        if not parsed:
            return None

        command_name, args = parsed
        return self.expand(command_name, args)

    def is_command(self, input_str: str) -> bool:
        """Check if input is a slash command.

        Args:
            input_str: Input to check

        Returns:
            True if input starts with /
        """
        return input_str.strip().startswith("/")



"""Tests for multi-agent patterns."""

import uuid

import pytest

from miu_core.models import Response, TextContent
from miu_core.patterns import (
    Orchestrator,
    OrchestratorConfig,
    Pipeline,
    PipelineConfig,
    Router,
    RouterConfig,
)


class MockAgent:
    """Mock agent for testing."""

    def __init__(self, response_text: str = "mock response") -> None:
        self.response_text = response_text
        self.run_count = 0
        self.last_query: str | None = None

    async def run(self, query: str) -> Response:
        """Mock run method."""
        self.run_count += 1
        self.last_query = query
        return Response(
            id=str(uuid.uuid4()),
            content=[TextContent(type="text", text=f"{self.response_text}: {query[:50]}")],
        )


class FailingAgent:
    """Agent that raises exceptions."""

    async def run(self, query: str) -> Response:
        """Always fails."""
        raise RuntimeError("Agent failed")


# =============================================================================
# Orchestrator Tests
# =============================================================================


class TestOrchestrator:
    """Tests for Orchestrator pattern."""

    @pytest.fixture
    def orchestrator(self) -> Orchestrator:
        """Create orchestrator instance."""
        return Orchestrator()

    async def test_single_task(self, orchestrator: Orchestrator) -> None:
        """Test single task execution."""
        agent = MockAgent("result")
        orchestrator.add_agent("test", agent)
        orchestrator.add_task("task1", "test", "do something")

        results = await orchestrator.run()

        assert "task1" in results
        assert results["task1"].success
        assert agent.run_count == 1
        assert "do something" in agent.last_query  # type: ignore

    async def test_task_with_dependencies(self, orchestrator: Orchestrator) -> None:
        """Test tasks with dependencies execute in order."""
        agent1 = MockAgent("first")
        agent2 = MockAgent("second")

        orchestrator.add_agent("agent1", agent1)
        orchestrator.add_agent("agent2", agent2)

        orchestrator.add_task("task1", "agent1", "first task")
        orchestrator.add_task(
            "task2",
            "agent2",
            lambda ctx: f"second task using {ctx['task1'].response.get_text()[:10]}",
            depends_on=["task1"],
        )

        results = await orchestrator.run()

        assert results["task1"].success
        assert results["task2"].success
        assert agent1.run_count == 1
        assert agent2.run_count == 1
        # Verify task2 received context from task1
        assert "first" in agent2.last_query  # type: ignore

    async def test_fail_fast(self) -> None:
        """Test fail_fast stops on error."""
        orchestrator = Orchestrator(OrchestratorConfig(fail_fast=True))

        orchestrator.add_agent("failing", FailingAgent())
        orchestrator.add_agent("second", MockAgent())

        orchestrator.add_task("task1", "failing", "will fail")
        orchestrator.add_task("task2", "second", "should not run", depends_on=["task1"])

        results = await orchestrator.run()

        assert not results["task1"].success
        assert "Agent failed" in (results["task1"].error or "")
        # Task2 should not exist or show dependency failure
        assert "task2" not in results or not results["task2"].success

    async def test_unregistered_agent_raises(self, orchestrator: Orchestrator) -> None:
        """Test adding task with unregistered agent."""
        with pytest.raises(ValueError, match="not registered"):
            orchestrator.add_task("task", "unknown_agent", "query")

    async def test_circular_dependency_detection(self, orchestrator: Orchestrator) -> None:
        """Test circular dependencies are detected."""
        agent = MockAgent()
        orchestrator.add_agent("a", agent)

        orchestrator.add_task("task1", "a", "q1", depends_on=["task2"])
        orchestrator.add_task("task2", "a", "q2", depends_on=["task1"])

        with pytest.raises(ValueError, match="Circular dependency"):
            await orchestrator.run()

    async def test_clear(self, orchestrator: Orchestrator) -> None:
        """Test clearing orchestrator."""
        orchestrator.add_agent("test", MockAgent())
        orchestrator.add_task("task", "test", "query")
        orchestrator.clear()

        results = await orchestrator.run()
        assert len(results) == 0


# =============================================================================
# Pipeline Tests
# =============================================================================


class TestPipeline:
    """Tests for Pipeline pattern."""

    @pytest.fixture
    def pipeline(self) -> Pipeline:
        """Create pipeline instance."""
        return Pipeline()

    async def test_single_stage(self, pipeline: Pipeline) -> None:
        """Test single stage pipeline."""
        agent = MockAgent("stage1")
        pipeline.add_stage("stage1", agent)

        result = await pipeline.run("initial query")

        assert result.success
        assert result.stages_completed == 1
        assert result.final_response is not None
        assert "stage1" in result.final_response.get_text()

    async def test_multiple_stages(self, pipeline: Pipeline) -> None:
        """Test pipeline with multiple stages."""
        agent1 = MockAgent("first")
        agent2 = MockAgent("second")

        pipeline.add_stage("s1", agent1).add_stage("s2", agent2)

        result = await pipeline.run("start")

        assert result.success
        assert result.stages_completed == 2
        assert len(result.stage_responses) == 2
        assert "s1" in result.stage_responses
        assert "s2" in result.stage_responses

    async def test_transform_function(self, pipeline: Pipeline) -> None:
        """Test transform between stages."""
        agent1 = MockAgent("data")
        agent2 = MockAgent("processed")

        pipeline.add_stage("extract", agent1)
        pipeline.add_stage(
            "transform",
            agent2,
            transform=lambda q, r: f"TRANSFORM: {r.get_text()}",
        )

        result = await pipeline.run("input")

        assert result.success
        assert agent2.last_query is not None
        assert "TRANSFORM" in agent2.last_query

    async def test_stop_on_error(self) -> None:
        """Test pipeline stops on error when configured."""
        pipeline = Pipeline(PipelineConfig(stop_on_error=True))

        pipeline.add_stage("s1", MockAgent("ok"))
        pipeline.add_stage("s2", FailingAgent())
        pipeline.add_stage("s3", MockAgent("never"))

        result = await pipeline.run("start")

        assert not result.success
        assert result.stages_completed == 1
        assert result.failed_stage == "s2"
        assert "Agent failed" in (result.error or "")

    async def test_continue_on_error(self) -> None:
        """Test pipeline continues on error when configured."""
        pipeline = Pipeline(PipelineConfig(stop_on_error=False))

        agent1 = MockAgent("first")
        agent3 = MockAgent("third")

        pipeline.add_stage("s1", agent1)
        pipeline.add_stage("s2", FailingAgent())
        pipeline.add_stage("s3", agent3)

        result = await pipeline.run("start")

        # Should complete despite error (2 successful stages)
        assert result.success
        assert result.stages_completed == 3

    async def test_stages_property(self, pipeline: Pipeline) -> None:
        """Test stages property returns names."""
        pipeline.add_stage("a", MockAgent())
        pipeline.add_stage("b", MockAgent())
        pipeline.add_stage("c", MockAgent())

        assert pipeline.stages == ["a", "b", "c"]
        assert len(pipeline) == 3


# =============================================================================
# Router Tests
# =============================================================================


class TestRouter:
    """Tests for Router pattern."""

    @pytest.fixture
    def router(self) -> Router:
        """Create router instance."""
        return Router()

    async def test_keyword_routing(self, router: Router) -> None:
        """Test routing by keywords."""
        code_agent = MockAgent("code response")
        write_agent = MockAgent("write response")

        router.add_route("code", code_agent, keywords=["python", "code", "debug"])
        router.add_route("writing", write_agent, keywords=["write", "essay"])

        result = await router.route("help me debug python code")

        assert result.agent_name == "code"
        assert "code response" in result.response.get_text()

    async def test_pattern_routing(self, router: Router) -> None:
        """Test routing by regex pattern."""
        math_agent = MockAgent("math")

        router.add_route("math", math_agent, pattern=r"\d+\s*[\+\-\*\/]\s*\d+")

        result = await router.route("calculate 5 + 3")

        assert result.agent_name == "math"

    async def test_condition_routing(self, router: Router) -> None:
        """Test routing by custom condition."""
        long_agent = MockAgent("long handler")

        router.add_route("long", long_agent, condition=lambda q: len(q) > 50)

        result = await router.route("a" * 60)

        assert result.agent_name == "long"

    async def test_priority_ordering(self, router: Router) -> None:
        """Test routes are checked by priority."""
        specific = MockAgent("specific")
        general = MockAgent("general")

        router.add_route("general", general, condition=lambda q: True, priority=0)
        router.add_route(
            "specific",
            specific,
            keywords=["special"],
            priority=10,  # Higher priority
        )

        result = await router.route("special request")

        assert result.agent_name == "specific"

    async def test_default_agent(self) -> None:
        """Test default agent fallback."""
        router = Router(RouterConfig(default_agent="fallback"))
        fallback = MockAgent("fallback")

        router.add_route("fallback", fallback, condition=lambda q: False)  # Never matches

        result = await router.route("any query")

        assert result.agent_name == "fallback"

    async def test_no_match_raises(self, router: Router) -> None:
        """Test unmatched query raises error."""
        router.add_route("specific", MockAgent(), keywords=["specific"])

        with pytest.raises(ValueError, match="Unable to route"):
            await router.route("unmatched query")

    def test_get_route_without_execution(self, router: Router) -> None:
        """Test get_route returns name without running agent."""
        agent = MockAgent()
        router.add_route("test", agent, keywords=["test"])

        route = router.get_route("test query")

        assert route == "test"
        assert agent.run_count == 0  # Agent not called

    def test_routes_property(self, router: Router) -> None:
        """Test routes property returns names."""
        router.add_route("a", MockAgent(), keywords=["a"])
        router.add_route("b", MockAgent(), keywords=["b"])

        # Routes sorted by priority (same priority = insertion order)
        assert set(router.routes) == {"a", "b"}

    def test_must_provide_routing_rule(self, router: Router) -> None:
        """Test that at least one routing rule is required."""
        with pytest.raises(ValueError, match="Must provide"):
            router.add_route("invalid", MockAgent())



"""Base tool interface."""

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel


class ToolResult(BaseModel):
    """Result from tool execution."""

    output: str
    success: bool = True
    error: str | None = None


class ToolContext(BaseModel):
    """Execution context for tools."""

    working_dir: str = "."
    session_id: str = "default"


class Tool(ABC):
    """Abstract base class for tools."""

    name: str
    description: str

    @abstractmethod
    def get_input_schema(self) -> type[BaseModel]:
        """Return Pydantic model for tool input."""
        ...

    @abstractmethod
    async def execute(self, ctx: ToolContext, **kwargs: Any) -> ToolResult:
        """Execute the tool with given inputs."""
        ...

    def to_schema(self) -> dict[str, Any]:
        """Convert tool to LLM-compatible schema."""
        return {
            "name": self.name,
            "description": self.description,
            "input_schema": self.get_input_schema().model_json_schema(),
        }



"""Command schema definitions."""

from pathlib import Path

from pydantic import BaseModel, Field


class Command(BaseModel):
    """Slash command definition."""

    name: str = Field(description="Command name without leading /")
    description: str = Field(default="", description="Short description of the command")
    argument_hint: str = Field(
        default="", alias="argument-hint", description="Hint for expected arguments"
    )
    content: str = Field(description="Command template content")
    source_path: Path | None = Field(default=None, description="Path to source file")

    def expand(self, arguments: str) -> str:
        """Expand command template with arguments.

        Args:
            arguments: Arguments to substitute

        Returns:
            Expanded command content
        """
        return self.content.replace("$ARGUMENTS", arguments)


def parse_command_file(content: str) -> tuple[dict[str, str], str]:
    """Parse a command markdown file with frontmatter.

    Args:
        content: File content

    Returns:
        Tuple of (frontmatter dict, body content)
    """
    frontmatter: dict[str, str] = {}
    body = content

    if content.startswith("---"):
        parts = content.split("---", 2)
        if len(parts) >= 3:
            # Parse frontmatter
            for line in parts[1].strip().split("\n"):
                if ":" in line:
                    key, value = line.split(":", 1)
                    frontmatter[key.strip()] = value.strip()
            body = parts[2].strip()

    return frontmatter, body



"""Session logger for debugging and replay."""

from datetime import datetime
from pathlib import Path
from typing import Any

from miu_core.logging.types import LogEntry, LogEventType


class SessionLogger:
    """Log session interactions for debugging and replay."""

    def __init__(self, save_dir: Path | None = None) -> None:
        """Initialize session logger.

        Args:
            save_dir: Directory to save log files
        """
        self.save_dir = save_dir or Path.cwd() / ".miu" / "logs"
        self._entries: list[LogEntry] = []
        self._session_id: str = ""
        self._active = False

    def start_session(self, session_id: str | None = None) -> str:
        """Start a new logging session.

        Args:
            session_id: Optional session ID, generates if not provided

        Returns:
            The session ID
        """
        if session_id is None:
            session_id = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
        self._session_id = session_id
        self._entries = []
        self._active = True

        self.log(LogEventType.SESSION_START, f"Session started: {session_id}")
        return session_id

    def end_session(self) -> None:
        """End the current session."""
        if self._active:
            self.log(LogEventType.SESSION_END, "Session ended")
            self._active = False

    def log(
        self,
        event_type: LogEventType,
        content: str = "",
        **metadata: Any,
    ) -> LogEntry:
        """Log an event.

        Args:
            event_type: Type of event
            content: Event content/message
            **metadata: Additional metadata

        Returns:
            Created log entry
        """
        entry = LogEntry(
            timestamp=datetime.utcnow(),
            event_type=event_type,
            content=content,
            metadata=metadata,
            session_id=self._session_id,
        )
        self._entries.append(entry)
        return entry

    def log_user_message(self, message: str) -> LogEntry:
        """Log a user message."""
        return self.log(LogEventType.USER_MESSAGE, message)

    def log_assistant_message(self, message: str) -> LogEntry:
        """Log an assistant message."""
        return self.log(LogEventType.ASSISTANT_MESSAGE, message)

    def log_tool_call(
        self,
        tool_name: str,
        tool_input: dict[str, Any],
    ) -> LogEntry:
        """Log a tool call."""
        return self.log(
            LogEventType.TOOL_CALL,
            f"Calling tool: {tool_name}",
            tool_name=tool_name,
            tool_input=tool_input,
        )

    def log_tool_result(
        self,
        tool_name: str,
        result: str,
        success: bool = True,
    ) -> LogEntry:
        """Log a tool result."""
        return self.log(
            LogEventType.TOOL_RESULT,
            result,
            tool_name=tool_name,
            success=success,
        )

    def log_error(self, error: str, **metadata: Any) -> LogEntry:
        """Log an error."""
        return self.log(LogEventType.ERROR, error, **metadata)

    def save(self) -> Path:
        """Save session log to file.

        Returns:
            Path to saved file
        """
        self.save_dir.mkdir(parents=True, exist_ok=True)
        filepath = self.save_dir / f"session_{self._session_id}.jsonl"

        with open(filepath, "w", encoding="utf-8") as f:
            for entry in self._entries:
                f.write(entry.model_dump_json() + "\n")

        return filepath

    @classmethod
    def load(cls, filepath: Path) -> list[LogEntry]:
        """Load session log from file.

        Args:
            filepath: Path to log file

        Returns:
            List of log entries
        """
        entries: list[LogEntry] = []
        with open(filepath, encoding="utf-8") as f:
            for line in f:
                if line.strip():
                    entries.append(LogEntry.model_validate_json(line))
        return entries

    @classmethod
    def replay(cls, filepath: Path) -> list[LogEntry]:
        """Load session log for replay.

        This is an alias for load() for semantic clarity.

        Args:
            filepath: Path to log file

        Returns:
            List of log entries
        """
        return cls.load(filepath)

    @property
    def session_id(self) -> str:
        """Get current session ID."""
        return self._session_id

    @property
    def entries(self) -> list[LogEntry]:
        """Get all log entries."""
        return self._entries.copy()

    @property
    def is_active(self) -> bool:
        """Check if session is active."""
        return self._active



"""Logging type definitions."""

from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field


class LogEventType(str, Enum):
    """Types of log events."""

    SESSION_START = "session_start"
    SESSION_END = "session_end"
    USER_MESSAGE = "user_message"
    ASSISTANT_MESSAGE = "assistant_message"
    TOOL_CALL = "tool_call"
    TOOL_RESULT = "tool_result"
    ERROR = "error"
    SYSTEM = "system"


class LogEntry(BaseModel):
    """A single log entry."""

    timestamp: datetime = Field(default_factory=datetime.utcnow)
    event_type: LogEventType
    content: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)
    session_id: str | None = None

    class Config:
        """Pydantic config."""

        use_enum_values = True



"""Base skill interface."""

from pathlib import Path

from pydantic import BaseModel


class Skill(BaseModel):
    """A skill that extends agent capabilities."""

    name: str
    description: str
    instructions: str
    scripts: list[Path] = []
    resources: list[Path] = []
    path: Path | None = None

    def to_prompt(self) -> str:
        """Convert skill to system prompt fragment."""
        parts = [f"## {self.name}\n"]
        if self.description:
            parts.append(f"{self.description}\n")
        if self.instructions:
            parts.append(f"\n{self.instructions}\n")
        if self.scripts:
            parts.append("\n### Available Scripts")
            for script in self.scripts:
                parts.append(f"- `{script.name}`: {script}")
            parts.append("")
        return "\n".join(parts)



"""Anthropic Claude provider."""

from typing import Any

from miu_core.models import (
    Message,
    Response,
    TextContent,
    ToolUseContent,
    Usage,
)
from miu_core.providers.base import LLMProvider, ToolSchema
from miu_core.tracing import Tracer, get_tracer
from miu_core.tracing.types import SpanAttributes

try:
    from anthropic import AsyncAnthropic
except ImportError as e:
    raise ImportError("anthropic package required. Install with: uv add miu-core[anthropic]") from e


class AnthropicProvider(LLMProvider):
    """Anthropic Claude LLM provider."""

    name = "anthropic"

    def __init__(self, model: str = "claude-sonnet-4-20250514") -> None:
        self.model = model
        self._client = AsyncAnthropic()
        self._tracer: Tracer = get_tracer("miu.provider")

    async def complete(
        self,
        messages: list[Message],
        tools: list[ToolSchema] | list[dict[str, Any]] | None = None,
        system: str | None = None,
        max_tokens: int = 4096,
    ) -> Response:
        """Send messages to Claude and get response."""
        with self._tracer.start_as_current_span("provider.complete") as span:
            span.set_attribute(SpanAttributes.PROVIDER_NAME, self.name)
            span.set_attribute(SpanAttributes.PROVIDER_MODEL, self.model)

            api_messages = self._convert_messages(messages)

            kwargs: dict[str, Any] = {
                "model": self.model,
                "messages": api_messages,
                "max_tokens": max_tokens,
            }
            if system:
                kwargs["system"] = system
            if tools:
                kwargs["tools"] = tools

            response = await self._client.messages.create(**kwargs)
            result = self._convert_response(response)

            # Record token usage and stop reason
            if result.usage:
                span.set_attribute(SpanAttributes.PROVIDER_TOKENS_INPUT, result.usage.input_tokens)
                span.set_attribute(
                    SpanAttributes.PROVIDER_TOKENS_OUTPUT, result.usage.output_tokens
                )
            span.set_attribute(SpanAttributes.PROVIDER_STOP_REASON, result.stop_reason)

            return result

    def _convert_messages(self, messages: list[Message]) -> list[dict[str, Any]]:
        """Convert internal messages to Anthropic format."""
        result: list[dict[str, Any]] = []
        for msg in messages:
            if msg.role == "system":
                continue  # System handled separately

            if isinstance(msg.content, str):
                result.append({"role": msg.role, "content": msg.content})
            else:
                content_blocks: list[dict[str, Any]] = []
                for block in msg.content:
                    if isinstance(block, TextContent):
                        content_blocks.append({"type": "text", "text": block.text})
                    elif isinstance(block, ToolUseContent):
                        content_blocks.append(
                            {
                                "type": "tool_use",
                                "id": block.id,
                                "name": block.name,
                                "input": block.input,
                            }
                        )
                    else:
                        content_blocks.append(
                            {
                                "type": "tool_result",
                                "tool_use_id": block.tool_use_id,
                                "content": block.content,
                                "is_error": block.is_error,
                            }
                        )
                result.append({"role": msg.role, "content": content_blocks})

        return result

    def _convert_response(self, response: Any) -> Response:
        """Convert Anthropic response to internal format."""
        content: list[TextContent | ToolUseContent] = []

        for block in response.content:
            if block.type == "text":
                content.append(TextContent(text=block.text))
            elif block.type == "tool_use":
                content.append(
                    ToolUseContent(
                        id=block.id,
                        name=block.name,
                        input=block.input,
                    )
                )

        return Response(
            id=response.id,
            content=content,
            stop_reason=response.stop_reason or "end_turn",
            usage=Usage(
                input_tokens=response.usage.input_tokens,
                output_tokens=response.usage.output_tokens,
            ),
        )



"""Session logging system."""

from miu_core.logging.session_logger import SessionLogger
from miu_core.logging.types import LogEntry, LogEventType

__all__ = [
    "LogEntry",
    "LogEventType",
    "SessionLogger",
]



"""Truncation strategies for memory management."""

from enum import Enum

from miu_core.models import Message


class TruncationStrategy(str, Enum):
    """Strategy for truncating conversation history."""

    FIFO = "fifo"  # First In, First Out - remove oldest messages
    SLIDING = "sliding"  # Keep first + last N messages
    SUMMARIZE = "summarize"  # Placeholder for future summarization


def estimate_tokens(message: Message) -> int:
    """Estimate token count for a message.

    Uses ~4 chars per token as rough estimate.
    """
    content = message.content
    if isinstance(content, str):
        text = content
    else:
        text = " ".join(block.text if hasattr(block, "text") else str(block) for block in content)
    # Rough estimate: ~4 chars per token
    return len(text) // 4 + 1


def truncate_fifo(messages: list[Message], max_tokens: int) -> tuple[list[Message], int]:
    """Truncate using FIFO strategy.

    Removes oldest messages (except first user message) until under limit.

    Returns: (truncated messages, tokens removed)
    """
    if not messages:
        return messages, 0

    total_tokens = sum(estimate_tokens(m) for m in messages)
    if total_tokens <= max_tokens:
        return messages, 0

    # Keep first message (usually important context)
    result = [messages[0]]
    tokens_kept = estimate_tokens(messages[0])

    # Add messages from end until we hit limit
    for msg in reversed(messages[1:]):
        msg_tokens = estimate_tokens(msg)
        if tokens_kept + msg_tokens <= max_tokens:
            result.insert(1, msg)
            tokens_kept += msg_tokens
        else:
            break

    tokens_removed = total_tokens - tokens_kept
    return result, tokens_removed


def truncate_sliding(
    messages: list[Message], max_tokens: int, keep_first: int = 1, keep_last: int = 10
) -> tuple[list[Message], int]:
    """Truncate using sliding window strategy.

    Keeps first N and last M messages.

    Returns: (truncated messages, tokens removed)
    """
    if len(messages) <= keep_first + keep_last:
        return messages, 0

    total_tokens = sum(estimate_tokens(m) for m in messages)

    # Keep first and last messages
    first = messages[:keep_first]
    last = messages[-keep_last:]
    result = first + last

    tokens_kept = sum(estimate_tokens(m) for m in result)
    tokens_removed = total_tokens - tokens_kept

    return result, tokens_removed



"""Orchestrator pattern for coordinating multiple agents."""

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

from pydantic import BaseModel

from miu_core.agents.base import Agent
from miu_core.models import Response


class OrchestratorConfig(BaseModel):
    """Configuration for orchestrator."""

    max_parallel: int = 5
    fail_fast: bool = True
    timeout: float | None = None


@dataclass
class TaskResult:
    """Result from an orchestrated task."""

    agent_name: str
    success: bool
    response: Response | None = None
    error: str | None = None


@dataclass
class Task:
    """A task to be executed by an agent."""

    name: str
    agent: Agent
    query: str | Callable[[dict[str, Any]], str]
    depends_on: list[str] = field(default_factory=list)


class Orchestrator:
    """Coordinates multiple agents to complete complex tasks.

    The orchestrator manages agent dependencies and execution order,
    allowing for both sequential and parallel execution patterns.

    Example:
        orchestrator = Orchestrator()
        orchestrator.add_agent("researcher", research_agent)
        orchestrator.add_agent("writer", writer_agent)

        orchestrator.add_task("research", "researcher", "Research AI agents")
        orchestrator.add_task(
            "write",
            "writer",
            lambda ctx: f"Write about: {ctx['research'].response.get_text()}",
            depends_on=["research"],
        )

        results = await orchestrator.run()
    """

    def __init__(self, config: OrchestratorConfig | None = None) -> None:
        self.config = config or OrchestratorConfig()
        self._agents: dict[str, Agent] = {}
        self._tasks: list[Task] = []

    def add_agent(self, name: str, agent: Agent) -> None:
        """Register an agent with a name."""
        self._agents[name] = agent

    def add_task(
        self,
        name: str,
        agent_name: str,
        query: str | Callable[[dict[str, Any]], str],
        depends_on: list[str] | None = None,
    ) -> None:
        """Add a task to be executed.

        Args:
            name: Unique task identifier
            agent_name: Name of the agent to execute this task
            query: Either a static query string or a callable that receives
                   the context dict and returns a query string
            depends_on: List of task names that must complete before this task
        """
        if agent_name not in self._agents:
            raise ValueError(f"Agent '{agent_name}' not registered")

        self._tasks.append(
            Task(
                name=name,
                agent=self._agents[agent_name],
                query=query,
                depends_on=depends_on or [],
            )
        )

    async def run(self) -> dict[str, TaskResult]:
        """Execute all tasks respecting dependencies.

        Returns:
            Dictionary mapping task names to their results
        """
        results: dict[str, TaskResult] = {}
        completed: set[str] = set()
        context: dict[str, Any] = {}

        # Build dependency graph
        task_map = {task.name: task for task in self._tasks}

        # Topological sort for execution order
        execution_order = self._topological_sort(task_map)

        for task_name in execution_order:
            task = task_map[task_name]

            # Check dependencies are satisfied
            for dep in task.depends_on:
                if dep not in completed:
                    raise RuntimeError(f"Dependency '{dep}' not satisfied for '{task_name}'")
                if self.config.fail_fast and not results[dep].success:
                    results[task_name] = TaskResult(
                        agent_name=task_name,
                        success=False,
                        error=f"Dependency '{dep}' failed",
                    )
                    continue

            # Build query from context if callable
            query = task.query(context) if callable(task.query) else task.query

            # Execute task
            try:
                response = await task.agent.run(query)
                result = TaskResult(
                    agent_name=task_name,
                    success=True,
                    response=response,
                )
            except Exception as e:
                result = TaskResult(
                    agent_name=task_name,
                    success=False,
                    error=str(e),
                )
                if self.config.fail_fast:
                    results[task_name] = result
                    break

            results[task_name] = result
            context[task_name] = result
            completed.add(task_name)

        return results

    def _topological_sort(self, task_map: dict[str, Task]) -> list[str]:
        """Sort tasks by dependencies (Kahn's algorithm)."""
        in_degree: dict[str, int] = {name: 0 for name in task_map}
        graph: dict[str, list[str]] = {name: [] for name in task_map}

        for name, task in task_map.items():
            for dep in task.depends_on:
                if dep in graph:
                    graph[dep].append(name)
                    in_degree[name] += 1

        # Start with tasks that have no dependencies
        queue = [name for name, degree in in_degree.items() if degree == 0]
        result: list[str] = []

        while queue:
            current = queue.pop(0)
            result.append(current)

            for dependent in graph[current]:
                in_degree[dependent] -= 1
                if in_degree[dependent] == 0:
                    queue.append(dependent)

        if len(result) != len(task_map):
            raise ValueError("Circular dependency detected in tasks")

        return result

    def clear(self) -> None:
        """Clear all agents and tasks."""
        self._agents.clear()
        self._tasks.clear()
